<?php declare(strict_types=1);

namespace Lubian\AttributeMagic\Infrastructure\WebApp\Route;

use FastRoute\RouteCollector;
use Lubian\AttributeMagic\Infrastructure\Event\AsListener;
use Lubian\AttributeMagic\Infrastructure\Event\Dispatcher;
use Lubian\AttributeMagic\Infrastructure\Route\Handler;
use Lubian\AttributeMagic\Infrastructure\Route\HttpMethod;
use Lubian\AttributeMagic\Infrastructure\WebApp\Request\RequestEvent;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;

use function count;
use function FastRoute\cachedDispatcher;
use function is_array;
use function is_string;

final readonly class HandlerResolver
{
    public function __construct(
        private Dispatcher $dispatcher,
        private bool $cached = false,
    ) {}

    #[AsListener(RequestEvent::class, -90)]
    public function resolveHandler(RequestEvent $event): void
    {
        $routeInfo = $this->getRouteInfo($event->request);

        if ($routeInfo[0] === \FastRoute\Dispatcher::NOT_FOUND) {
            $this->notFound($event);
            return;
        }

        if ($routeInfo[0] === \FastRoute\Dispatcher::METHOD_NOT_ALLOWED) {
            $event->response = new Response('Not Allowed', 403);
            $event->stopped = true;
            return;
        }

        if (
            is_array($routeInfo[1]) === false
            || count($routeInfo[1]) !== 2
            || is_string($routeInfo[1][0]) === false
            || $routeInfo[1][0] === ''
            || is_string($routeInfo[1][1]) === false
            || $routeInfo[1][1] === ''
        ) {
            $this->notFound($event);
            return;
        }

        $event->handler = new Handler(
            HttpMethod::from($event->request->getMethod()),
            $event->request->getPathInfo(),
            $routeInfo[1][0],
            $routeInfo[1][1],
        );

        /** @var array<array-key, mixed> $args */
        $args = $routeInfo[2];
        $event->request->attributes->add($args);
    }

    private function getRouteInfo(Request $request): array
    {
        $dispatcher = $this->createRouteDispatcher();
        return $dispatcher->dispatch($request->getMethod(), $request->getPathInfo());
    }

    private function createRouteDispatcher(): \FastRoute\Dispatcher
    {
        $routesEvent = new CollectRoutes($this->cached);
        $this->dispatcher->dispatch($routesEvent);
        $dispatcher = $this->dispatcher;
        return cachedDispatcher(
            static function (RouteCollector $r) use ($dispatcher, $routesEvent): void {
                $dispatcher->dispatch($routesEvent);

                foreach ($routesEvent->routes as $h) {
                    $r->addRoute(
                        $h->method->value,
                        $h->path,
                        [$h->handlerClass, $h->handlerMethod],
                    );
                }
            },
            [
                'cacheFile' => __DIR__ . '/../../../../var/route.cache',
                'cacheDisabled' => ! $this->cached,
            ],
        );
    }

    private function notFound(RequestEvent $event): void
    {
        $event->response = new Response('Not Found', 404);
        $event->stopped = true;
    }
}
